from influxdb_client import NotificationEndpointsService, HTTPNotificationEndpoint, NotificationRulesService, \
    HTTPNotificationRule, StatusRule, RuleStatusLevel, TaskStatusType
from influxdb_client.rest import ApiException

from dms_influx2.exceptions import NotificationError


class NotificationEndpointApi(NotificationEndpointsService):
    def __init__(self, client):
        self.org = client.org
        self.org_id = client.organizations_api().get_id_organization_by_name(self.org)
        super().__init__(api_client=client.api_client)

    def get_all_notification_endpoints(self):
        return self.get_notification_endpoints(org_id=self.org_id)

    def notification_endpoints_id_exists(self, endpoint_id):
        try:
            self.get_notification_endpoints_id(endpoint_id)
            return True
        except ApiException as e:
            if e.status == 404 or e.status == 400:
                return False

    def add_http_notification(self, name, url, description=None):
        endpoint = HTTPNotificationEndpoint(name=name,
                                            url=url,
                                            org_id=self.org_id,
                                            method='POST',
                                            auth_method="none",
                                            description=description)
        service = NotificationEndpointsService(api_client=self.api_client)
        try:
            return service.create_notification_endpoint(endpoint)
        except ApiException as e:
            if e.status == 422:
                raise NotificationError(f"Notification endpoint with name '{name}' already exists.")


class NotificationRuleApi(NotificationRulesService):
    def __init__(self, client):
        self.org = client.org
        self.org_id = client.organizations_api().get_id_organization_by_name(self.org)
        super().__init__(api_client=client.api_client)

    def get_all_notification_rules(self):
        return self.get_notification_rules(org_id=self.org_id)

    def notification_rule_id_exists(self, rule_id):
        try:
            self.get_notification_rules_id(rule_id)
            return True
        except ApiException as e:
            if e.status == 404 or e.status == 400:
                return False

    def add_notification_rule(self, name, endpoint_id, status_rules=None):
        if status_rules is None:
            status_rules = [
                {
                    'count': None,
                    'current_level': 'ANY',
                    'period': None,
                    'previous_level': 'ANY'
                }
            ]
        notification_rule = HTTPNotificationRule(name=name,
                                                 every="5s",
                                                 offset="0s",
                                                 status_rules=[StatusRule(previous_level=RuleStatusLevel.ANY,
                                                                          current_level=RuleStatusLevel.ANY)],
                                                 tag_rules=[],
                                                 endpoint_id=endpoint_id,
                                                 org_id=self.org_id,
                                                 status=TaskStatusType.ACTIVE)

        service = NotificationRulesService(api_client=self.api_client)
        return service.create_notification_rule(notification_rule)



from typing import Union
from dateutil.parser import parse
import pytz
from datetime import datetime, timedelta


def timestamp_to_influx_string(timestamp: Union[str, datetime], offset: int = None) -> str:
    if type(timestamp) == str:
        timestamp = parse(timestamp)
    if offset is not None:
        timestamp = timestamp - timedelta(hours=int(offset))
    return timestamp.strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def localize_dt(dt: datetime, timezone: str = 'Europe/Ljubljana', to_str: bool = False):
    tz = pytz.timezone(timezone)
    dt_localized = tz.localize(dt)
    offset = dt_localized.utcoffset().total_seconds() / 60 / 60
    dt = dt_localized.replace(tzinfo=None) + timedelta(hours=offset)
    if to_str:
        dt = str(dt)
    return dt


def dt_to_utc(dt: datetime, timezone: str = 'Europe/Ljubljana', to_str: bool = False):
    tz = pytz.timezone(timezone)
    dt_localized = tz.localize(dt)
    offset = dt_localized.utcoffset().total_seconds() / 60 / 60
    dt = dt_localized.replace(tzinfo=None) - timedelta(hours=offset)
    if to_str:
        dt = str(dt)
    return dt


if __name__ == '__main__':
    print(localize_dt(datetime.now()))



import logging
from datetime import datetime

logger = logging.getLogger(__name__)


def parse_dtv_bytes(file: bytes) -> dict:
    """
    :param file: File that has dtv styling
    :return: {unit:<unit>, data:[<timestamp>,<value>]}
    """

    lines = file.decode().split('\n')

    # Retrieve unit
    unit = ''
    for line in lines[0:8]:
        item = line.strip()
        if item != '':
            unit = item

    # Retrieve data
    values = []
    for line in lines:
        item = line.split(',')
        if len(item) == 2:
            try:
                ts = item[0].strip()
                val = float(item[1].strip())
                if '.' in ts:
                    dt = datetime.strptime(ts, "%d/%m/%Y %H:%M:%S.%f")
                else:
                    dt = datetime.strptime(ts, "%d/%m/%Y %H:%M:%S")
            except ValueError as e:
                logger.exception(e)
                # Invalid datetime string
                continue
            values.append((dt, val))

    return {"unit": unit, "values": values}


def create_dtv_file(device, channel, unit, values):
    lines = ['\n', '\n', unit, '\n', '\n', '\n', '\n', '\n', '\n', '\n']
    filename = f'{device}_{channel}.dtv'
    for dt, value in values:
        ts = dt.replace(microsecond=0).strftime("%d/%m/%Y %H:%M:%S")
        lines.append(f'{ts},{value}\n')
    with open(filename, 'w') as f:
        f.writelines(lines)
    return filename



from influxdb_client import TasksApi


class Tasks(TasksApi):
    def __init__(self, client):
        self.org = client.org
        self.org_id = client.organizations_api().get_organization(org_name=self.org)
        super().__init__(client)

    def create_or_update_calibration_task(self,
                                          bucket,
                                          measurement,
                                          channel_src,
                                          channel_dest,
                                          const_a=None,
                                          const_b=None,
                                          const_c=None,
                                          const_d=None,
                                          offset=None,
                                          altitude=None,
                                          baro_url=None,
                                          baro_allowed_hours_diff=12,
                                          task_id=None):
        if bucket is None:
            raise ValueError('Value for argument `bucket` must not be None')
        if measurement is None:
            raise ValueError('Value for argument `measurement` must not be None')
        if channel_src is None:
            raise ValueError('Value for argument `channel_src` must not be None')
        if channel_dest is None:
            raise ValueError('Value for argument `channel_dest` must not be None')
        if const_a is None:
            const_a = 0
        if const_b is None:
            const_b = 0
        if const_c is None:
            const_c = 1
        if const_d is None:
            const_d = 0
        if offset is None:
            offset = 0
        if altitude is None:
            altitude = 0
        if baro_allowed_hours_diff is None:
            baro_allowed_hours_diff = 12

        const_a = float(const_a)
        const_b = float(const_b)
        const_c = float(const_c)
        const_d = float(const_d)
        offset = float(offset)
        altitude = float(altitude)

        name = f'calibration_{bucket}_{measurement}_{channel_src}_{channel_dest}'

        if baro_url and baro_allowed_hours_diff:
            header = f'response = requests.get(url: "{baro_url}", params: ["allowed_hours_diff": ["{baro_allowed_hours_diff}"]])\n'\
                     f'valueH = json.parse(data: response.body)["value_h"]\n' \
                     f'measurement = if response.statusCode == 200 then "{measurement}" else ""'
        else:
            header = f'measurement = "{measurement}"\n' \
                     f'valueH = 0.0'

        value = f'math.round(x: ({const_a} * math.pow(x: r._value, y: 3.0) + {const_b} * math.pow(x: r._value, y: 2.0) + ' \
                f'{const_c} * math.pow(x: r._value, y: 1.0) + {const_d} + {offset} + {altitude} - valueH) * 1000.0) / 1000.0'

        flux = f'{header}\n' \
               f'from(bucket: "{bucket}")\n' \
               f'|> range(start: -task.every)\n' \
               f'|> filter(fn: (r) => r._measurement == measurement)\n' \
               f'|> filter(fn: (r) => r["_field"] == "value")\n' \
               f'|> filter(fn: (r) => r["channel"] == "{channel_src}")\n' \
               f'|> map(fn: (r) => ({{r with _value: {value} }}))\n' \
               f'|> map(fn: (r) => ({{r with channel: "{channel_dest}", device_id: "${{r.device}}.{channel_dest}"}}))' \
               f'|> to(bucket: "{bucket}")'
        print(flux)
        if task_id is not None:
            task = self.find_task_by_id(task_id=task_id)
        else:
            task = self.create_task_every(name=name, flux=flux, organization=self.org_id, every='1m')

        task.flux = 'import "math"\n'\
                    'import "http/requests"\n' \
                    'import "experimental/json"\n'\
                    f'option task = {{name: "{name}", every: 1m}} \n' + flux
        return self.update_task(task)



import os
from time import sleep

import pytest

from dms_influx2.lib import DmsInflux2
from datetime import datetime, timedelta
from dateutil.parser import parse

from dms_influx2.utils import localize_dt

DATABASE_URL = os.environ['DATABASE_URL']
TOKEN = os.environ['TOKEN']
ORG = os.environ['ORG']
CLIENT = DmsInflux2(url=DATABASE_URL, token=TOKEN, org=ORG)

BUCKET = 'test_bucket_12345'
BUCKET_TRASH = f"{BUCKET}_trash_{datetime.now().year}"
MEASUREMENT = "test"
DEVICE_ID = "test.ch"
DESCRIPTION = "some test"
START = 0
N_SAMPLES = 10000
TEST_DATA = {
    "measurement": MEASUREMENT,
    "device_id": DEVICE_ID,
    "description": DESCRIPTION,
    "values": [(datetime.now() - timedelta(minutes=i), i) for i in range(START, N_SAMPLES)]
}


@pytest.fixture(scope='session')
def init():
    CLIENT.save_data(bucket=BUCKET, data=TEST_DATA)


def test_health():
    health = CLIENT.ping()
    assert health


def test_create_delete_bucket(init):
    bucket_name = 'abrakadabra'
    CLIENT.buckets_api().create_bucket(bucket_name=bucket_name)
    CLIENT.buckets_api().delete_bucket_by_name(bucket_name=bucket_name)
    assert CLIENT.buckets_api().bucket_exists(bucket_name) is False


def test_list_measurements(init):
    data = CLIENT.list_measurements(bucket=BUCKET)
    assert MEASUREMENT in data


def test_list_device_ids(init):
    data = CLIENT.list_device_ids(bucket=BUCKET)
    assert DEVICE_ID in data


def test_list_descriptions(init):
    data = CLIENT.list_descriptions(bucket=BUCKET, measurement=MEASUREMENT)
    assert data == [DESCRIPTION]

    data = CLIENT.list_descriptions(bucket=BUCKET)
    assert data == [DESCRIPTION]


def test_get_values_count_combined(init):
    data = CLIENT.get_values_count_combined(bucket=BUCKET, measurement=MEASUREMENT, device_id=DEVICE_ID)
    assert data == N_SAMPLES


def test_get_values_count(init):
    data = CLIENT.get_values_count(bucket=BUCKET, measurement=MEASUREMENT, device_id=DEVICE_ID)
    assert data[DEVICE_ID] == N_SAMPLES


def test_get_metadata(init):
    data = CLIENT.get_metadata(bucket=BUCKET, measurement=MEASUREMENT, device_id=DEVICE_ID)
    assert data[0]['device_id'] == DEVICE_ID
    assert data[0]['values_count'] == N_SAMPLES
    assert data[0]['timestamp'] == str(TEST_DATA['values'][0][0])
    assert data[0]['value'] == TEST_DATA['values'][0][1]


def test_get_first_value(init):
    data = CLIENT.get_one_value(bucket=BUCKET, measurement=MEASUREMENT, last=False)
    assert data[0]['value'] == TEST_DATA['values'][-1][1]


def test_get_last_value(init):
    data = CLIENT.get_one_value(bucket=BUCKET, measurement=MEASUREMENT, last=True)
    assert data[0]['value'] == START


def test_get_last_value_sort(init):
    # Get last value with sorting (asc or desc)
    d1 = {'measurement': 'test1', 'device_id': 'test.1', 'unit': 't',
          'values': [(datetime.now() - timedelta(hours=2), 0)]}
    d2 = {'measurement': 'test1', 'device_id': 'test.2', 'unit': 't',
          'values': [(datetime.now() - timedelta(hours=0), 0)]}
    d3 = {'measurement': 'test1', 'device_id': 'test.3', 'unit': 't',
          'values': [(datetime.now() - timedelta(hours=1), 0)]}
    CLIENT.save_data(bucket=BUCKET, data=[d1, d3, d2])

    # ascending sorting
    data = CLIENT.get_one_value(bucket=BUCKET, measurement='test1', sort='asc', last=True)
    assert data[0]['device_id'] == 'test.1'
    assert data[1]['device_id'] == 'test.3'
    assert data[2]['device_id'] == 'test.2'

    # descending sorting
    data = CLIENT.get_one_value(bucket=BUCKET, measurement='test1', sort='desc', last=True)
    assert data[0]['device_id'] == 'test.2'
    assert data[1]['device_id'] == 'test.3'
    assert data[2]['device_id'] == 'test.1'


def test_get_values_from_device_id(init):
    # Descending order
    data = CLIENT.get_values_from_device_id(bucket=BUCKET, measurement=MEASUREMENT, device_id=DEVICE_ID, sort='desc',
                                            time_range='all')
    assert data == list([(str(ts), val) for ts, val in TEST_DATA['values']])

    # Ascending order
    data = CLIENT.get_values_from_device_id(bucket=BUCKET, measurement=MEASUREMENT, device_id=DEVICE_ID, sort='asc',
                                            time_range='all')
    assert data == [(str(ts), val) for ts, val in reversed(TEST_DATA['values'])]

    # Test different time ranges
    values = CLIENT.get_values_from_device_id(bucket=BUCKET, measurement=MEASUREMENT, device_id=DEVICE_ID,
                                              time_range='all')
    assert len(values) == N_SAMPLES

    # Get values within 1 hour (60samples)
    values = CLIENT.get_values_from_device_id(bucket=BUCKET, measurement=MEASUREMENT, device_id=DEVICE_ID,
                                              time_range='1h')
    assert 60 >= len(values) > 58

    # Get values within 1 day (60*24)
    values = CLIENT.get_values_from_device_id(bucket=BUCKET, measurement=MEASUREMENT, device_id=DEVICE_ID,
                                              time_range='1d')
    assert 60 * 24 + 2 > len(values) > 60 * 24 - 2

    time_from = (datetime.utcnow() - timedelta(hours=10)).replace(minute=0)
    time_to = time_from + timedelta(hours=1)
    values = CLIENT.get_values_from_device_id(bucket=BUCKET, measurement=MEASUREMENT, device_id=DEVICE_ID,
                                              time_from=time_from, time_to=time_to)
    assert len(values) == 60

    time_from = (datetime.utcnow() - timedelta(hours=100)).replace(minute=0)
    time_to = time_from + timedelta(hours=5)
    values = CLIENT.get_values_from_device_id(bucket=BUCKET, measurement=MEASUREMENT, device_id=DEVICE_ID,
                                              time_from=time_from, time_to=time_to)
    assert len(values) == 60 * 5


def test_range_values(init):
    dt = datetime.utcnow()
    _measurement = 'valuerange'
    _device_id = 'valuerange.1'
    _data = {
        "measurement": _measurement,
        "device_id": _device_id,
        "values": [
            (datetime(2023, 1, 22, 12, 0, 0), 0),
            (datetime(2023, 1, 22, 13, 0, 0), 1),
            (datetime(2023, 1, 22, 14, 0, 0), 2),
            (datetime(2023, 1, 22, 15, 0, 0), 3)
        ]
    }
    CLIENT.save_data(bucket=BUCKET, data=_data)

    data = CLIENT.get_values_from_device_id(bucket=BUCKET,
                                            measurement=_measurement,
                                            device_id=_device_id,
                                            time_range='all',
                                            value_type='greater',
                                            value=2)
    assert len(data) == 1
    assert data[0][1] == _data['values'][-1][1]

    data = CLIENT.get_values_from_device_id(bucket=BUCKET,
                                            measurement=_measurement,
                                            device_id=_device_id,
                                            time_range='all',
                                            value_type='lesser',
                                            value=1)
    assert len(data) == 1
    assert data[0][1] == _data['values'][0][1]

    data = CLIENT.get_values_from_device_id(bucket=BUCKET,
                                            measurement=_measurement,
                                            device_id=_device_id,
                                            time_range='all',
                                            value_type='range',
                                            value_within=True,
                                            value_min=1,
                                            value_max=3)
    assert len(data) == 1
    assert data[0][1] == _data['values'][2][1]

    data = CLIENT.get_values_from_device_id(bucket=BUCKET,
                                            measurement=_measurement,
                                            device_id=_device_id,
                                            time_range='all',
                                            value_type='range',
                                            value_within=False,
                                            value_min=1,
                                            value_max=2)
    assert len(data) == 2
    assert data[0][1] == _data['values'][-1][1]
    assert data[1][1] == _data['values'][0][1]


def test_timezones(init):
    dt = datetime.utcnow()
    _measurement = 'timezone'
    _device_id = 'timezone.1'
    data = {
        "measurement": _measurement,
        "device_id": _device_id,
        "description": DESCRIPTION,
        "values": [(dt, 1)]
    }
    CLIENT.save_data(bucket=BUCKET, data=data)

    cl2 = DmsInflux2(url=DATABASE_URL, token=TOKEN, org=ORG)
    data = cl2.get_one_value(bucket=BUCKET, measurement=_measurement, device_id=_device_id, time_range='all')
    assert parse(data[0]['timestamp']) == dt

    cl2 = DmsInflux2(url=DATABASE_URL, token=TOKEN, org=ORG, timezone_offset=1)
    data = cl2.get_one_value(bucket=BUCKET, measurement=_measurement, device_id=_device_id, time_range='all')
    assert parse(data[0]['timestamp']) == dt + timedelta(hours=1)

    cl2 = DmsInflux2(url=DATABASE_URL, token=TOKEN, org=ORG, timezone_offset=-1)
    data = cl2.get_one_value(bucket=BUCKET, measurement=_measurement, device_id=_device_id, time_range='all')
    assert parse(data[0]['timestamp']) == dt + timedelta(hours=-1)


def test_utc_conversion(init):
    dt1 = datetime(2022, 3, 25, 12, 0, 0)
    dt2 = datetime(2022, 3, 27, 12, 0, 0)
    _measurement = 'tt'
    _device_id = 'tt.1'
    data = {
        "measurement": _measurement,
        "device_id": _device_id,
        "values": [(dt1, 1), (dt2, 2)]
    }
    CLIENT.save_data(bucket=BUCKET, data=data, utc_to_local=True)
    data = CLIENT.get_values_from_device_id(bucket=BUCKET,
                                            measurement=_measurement,
                                            device_id=_device_id,
                                            time_range='all')
    assert data[0][0] == localize_dt(dt2, to_str=True)
    assert data[-1][0] == localize_dt(dt1, to_str=True)


def test_delete_data():
    bucket = 'd_buck'
    test_data = {
        "measurement": 'deletetss',
        "device_id": 'deletetss.12',
        "values": [
            (datetime(2023, 1, 1, 12, 0, 0), 0),
            (datetime(2023, 1, 1, 12, 1, 0), 0),
            (datetime(2023, 1, 1, 12, 2, 0), 0),
            (datetime(2023, 1, 1, 12, 3, 0), 0),
            (datetime.now(), 0),
        ]
    }
    CLIENT.save_data(bucket=bucket, data=test_data)

    # Test delete time_range (from -> to)
    CLIENT.delete_api().delete_data(bucket=bucket,
                                    measurements=[test_data['measurement']],
                                    time_from=datetime(2023, 1, 1, 12, 0, 0),
                                    time_to=datetime(2023, 1, 1, 12, 2, 0))
    data = CLIENT.get_data_from_device_id(bucket=bucket,
                                          measurement=test_data['measurement'],
                                          device_id=test_data['device_id'],
                                          time_range='all')
    assert len(data['values']) == 2

    # Test delete all
    CLIENT.delete_api().delete_data(bucket=bucket,
                                    measurements=[test_data['measurement']])
    data = CLIENT.get_data_from_device_id(bucket=bucket,
                                          measurement=test_data['measurement'],
                                          device_id=test_data['device_id'],
                                          time_range='all')
    assert data == {}


def test_downsample_data():
    bucket = 'downsample_test'
    values = [(str(datetime.now() - timedelta(hours=i)), i) for i in range(5, 100)]
    test_data = [
        {
            "measurement": 'downsample',
            "device_id": 'downsample.12',
            "values": values
        },
        {
            "measurement": 'downsample',
            "device_id": 'downsample.123',
            "values": values
        }
    ]
    CLIENT.save_data(bucket=bucket, data=test_data)
    '''Test time_from and time_to'''
    time_from = parse(test_data[0]['values'][-1][0]) - timedelta(seconds=1)
    time_to = parse(test_data[0]['values'][0][0]) + timedelta(seconds=1)
    aggregate_window = '30m'
    aggregated_data = CLIENT.get_data_from_device_id(bucket=bucket,
                                                     measurement=test_data[0]['measurement'],
                                                     device_id=test_data[0]['device_id'],
                                                     time_from=time_from,
                                                     time_to=time_to,
                                                     aggregate_window=aggregate_window)
    # Perform a downsample and check if all data are downsampled within bucket
    CLIENT.down_sample(bucket=bucket,
                       measurement=test_data[0]['measurement'],
                       device_id=test_data[0]['device_id'],
                       aggregate_window=aggregate_window,
                       time_from=time_from,
                       time_to=time_to)
    data_in_downsampled_bucket = CLIENT.get_data_from_device_id(bucket=CLIENT.downsample_temp_bucket,
                                                                measurement=test_data[0]['measurement'],
                                                                device_id=test_data[0]['device_id'],
                                                                time_range='all')
    assert aggregated_data['values'] == data_in_downsampled_bucket['values']
    # New data in this bucket must be equal to aggregated data
    new_data = CLIENT.get_data_from_device_id(bucket=bucket,
                                              measurement=test_data[0]['measurement'],
                                              device_id=test_data[0]['device_id'],
                                              time_from=time_from - timedelta(seconds=1),
                                              time_to=time_to + timedelta(seconds=1))
    assert aggregated_data['values'] == new_data['values']
    # New data for second device must be the same as origin
    new_data_second = CLIENT.get_data_from_device_id(bucket=bucket,
                                                     measurement=test_data[-1]['measurement'],
                                                     device_id=test_data[-1]['device_id'],
                                                     time_range='all')
    assert test_data[-1]['values'] == new_data_second['values']

    '''Test2: test older_than param'''
    bucket = 'downsample_test1'
    older_than = '1d'
    aggregate_window = '4h'
    CLIENT.save_data(bucket=bucket, data=test_data)
    time_from = datetime(1970, 1, 1, 12, 0, 0)
    time_to = datetime.now() - timedelta(days=1) + timedelta(seconds=1)
    aggregated_data = CLIENT.get_data_from_device_id(bucket=bucket,
                                                     measurement=test_data[0]['measurement'],
                                                     device_id=test_data[0]['device_id'],
                                                     time_from=time_from,
                                                     time_to=time_to,
                                                     aggregate_window=aggregate_window)
    # Perform a downsample and check if all data are downsampled within bucket
    CLIENT.down_sample(bucket=bucket,
                       measurement=test_data[0]['measurement'],
                       device_id=test_data[0]['device_id'],
                       aggregate_window=aggregate_window,
                       older_than=older_than)
    data_in_downsampled_bucket = CLIENT.get_data_from_device_id(bucket=CLIENT.downsample_temp_bucket,
                                                                measurement=test_data[0]['measurement'],
                                                                device_id=test_data[0]['device_id'],
                                                                time_range='all')
    assert len(aggregated_data['values']) == len(data_in_downsampled_bucket['values'])
    # Old data must remain the same
    old_data = CLIENT.get_data_from_device_id(bucket=bucket,
                                              measurement=test_data[0]['measurement'],
                                              device_id=test_data[0]['device_id'],
                                              time_from=time_to)
    old_data_values = []
    for value in values:
        if parse(value[0]) > (time_to - timedelta(seconds=1)):
            old_data_values.append((str(value[0]), value[1]))
    assert old_data_values == old_data['values']
    new_data = CLIENT.get_data_from_device_id(bucket=bucket,
                                              measurement=test_data[0]['measurement'],
                                              device_id=test_data[0]['device_id'],
                                              time_from=time_from,
                                              time_to=time_to)
    assert len(aggregated_data['values']) == len(new_data['values'])





from influxdb_client import AuthorizationsApi, Authorization

from dms_influx2.exceptions import BucketApiError


class Authorizations(AuthorizationsApi):
    def __init__(self, client):
        self.org = client.org
        self.organizations_api = client.organizations_api
        self.buckets_api = client.buckets_api
        super().__init__(client)

    def create_bucket_authorization(self, bucket_name, org_name=None, read_only=True, description=None):
        if org_name is None:
            org_name = self.org
        org = self.organizations_api().get_organization(org_name=org_name)
        bucket = self.buckets_api().find_bucket_by_name(bucket_name=bucket_name)

        permissions_read = {
            'action': 'read',
            'resource': {
                'id': bucket.id,
                'name': bucket.name,
                'org': org.name,
                'org_id': org.id,
                'type': 'buckets'
            }
        }
        permissions_write = {
            'action': 'write',
            'resource': {
                'id': bucket.id,
                'name': bucket.name,
                'org': org.name,
                'org_id': org.id,
                'type': 'buckets'
            }
        }
        permissions = [permissions_read]
        if not read_only:
            permissions.append(permissions_write)
        authorization = Authorization(org_id=org.id, permissions=permissions, description=description)
        return self.create_authorization(authorization=authorization)

    def create_single_bucket_authorization(self, bucket_name: str, description: str = None, org_name: str = None,
                                           write: bool = False, ) -> Authorization:
        if org_name is None:
            org_name = self.org

        if not self.buckets_api().bucket_exists(bucket_name=bucket_name):
            raise BucketApiError(f"Bucket '{bucket_name}' does not exists")

        return self.create_bucket_authorization(bucket_name=bucket_name,
                                                org_name=org_name,
                                                read_only=not write,
                                                description=description)

    def find_authorizations_by_description(self, description=None):
        auths = self.find_authorizations()
        auths_out = []
        for auth in auths:
            if auth.description == description:
                auths_out.append(auth)
        return auths_out

    def delete_all_bucket_authorizations(self):
        # Delete all authorizations

        for auth in self.find_authorizations():
            for perm in auth.permissions:
                if perm.resource.type != 'buckets':
                    break
            else:
                self.delete_authorization(auth)

    def get_or_create_bucket_authorization(self, bucket_name, read_only=True, active=True):
        # try authorization by name first
        # then name_readonly

        auth_description = f'{bucket_name}_readonly' if read_only else bucket_name

        # TODO
        # Try find authorization by description
        auth = self.find_authorizations_by_description(auth_description)
        if auth:
            auth = auth[0]
            for perm in auth.permissions:
                if perm.resource.type == 'buckets' and perm.resource.name == bucket_name:
                    return auth
        return self.create_bucket_authorization(bucket_name,
                                                org_name=None,
                                                read_only=read_only,
                                                description=auth_description)

    def __attach_bucket_auths_dict(self, auth):
        buckets_perm = {}
        for perm in auth.permissions:
            if perm.resource.type == 'buckets':
                if perm.resource.name not in buckets_perm:
                    buckets_perm[perm.resource.name] = []
                buckets_perm[perm.resource.name].append(perm.action)
        _auth = auth.to_dict()
        _auth['buckets'] = []
        for name, actions in buckets_perm.items():
            _auth['buckets'].append({
                'bucket_name': name,
                'read_only': 'write' not in actions
            })
        return _auth

    def get_bucket_authorization_dict(self, auth_id):
        auth = self.find_authorization_by_id(auth_id)
        return self.__attach_bucket_auths_dict(auth)

    def get_bucket_authorizations_dict(self, org_name=None, order_by_created=True):
        if org_name is None:
            org_name = self.org
        auths = self.find_authorizations_by_org_name(org_name=org_name)
        auths_out = []
        for auth in auths:
            _auth = self.__attach_bucket_auths_dict(auth)
            auths_out.append(_auth)
        if order_by_created:
            auths_out = sorted(auths_out, key=lambda b: b['created_at'], reverse=True)
        return auths_out










import os
import pytest
from influxdb_client.rest import ApiException

from dms_influx2.lib import DmsInflux2

DATABASE_URL = os.environ['DATABASE_URL']
TOKEN = os.environ['TOKEN']
ORG = os.environ['ORG']
BUCKET = 'test_bucket'
CLIENT = DmsInflux2(url=DATABASE_URL, token=TOKEN, org=ORG)


def test_bucket_exists():
    if CLIENT.buckets_api().bucket_exists(BUCKET):
        CLIENT.buckets_api().delete_bucket_by_name(BUCKET)
    CLIENT.buckets_api()._create_bucket(bucket_name=BUCKET)
    assert CLIENT.buckets_api().bucket_exists(BUCKET)


def test_delete_authorization():
    auth = CLIENT.authorizations_api().create_bucket_authorization(bucket_name=BUCKET, org_name=ORG)
    assert len(auth.id) > 5

    assert CLIENT.authorizations_api().delete_authorization(auth) is None

    with pytest.raises(ApiException):
        CLIENT.authorizations_api().find_authorization_by_id(auth_id=auth.id)

# # # TODO fix this test
# def test_create_bucket_with_token_and_delete_permissions():
#     """Test create new bucket with auth token and check for writes/reads permissions"""
#
#     new_token = client.buckets_api().create_bucket_with_auth(bucket_name=bucket).token
#
#     measurement = "test"
#
#     client2 = DmsInflux2(url=url, token=new_token, org=org)
#     data = {
#         "measurement": measurement,
#         "device_id": "test.ch",
#         "values": [(datetime.utcnow(), 0)]
#     }
#     client2.save_data(bucket=bucket, data=data)
#     client.delete_api().delete_data(bucket=bucket, measurements=[measurement])
#
#     client.buckets_api().delete_permissions(bucket_name=bucket)
#     data = {
#         "measurement": measurement,
#         "device_id": "test.ch",
#         "values": [(datetime.utcnow(), 0)]
#     }
#     with pytest.raises(Exception):
#         client2.save_data(bucket=bucket, data=data)



import os

from dms_influx2.lib import DmsInflux2
from datetime import datetime, timedelta
from dotenv import dotenv_values

DATABASE_URL = os.environ['DATABASE_URL']
TOKEN = os.environ['TOKEN']
ORG = os.environ['ORG']
CLIENT = DmsInflux2(url=DATABASE_URL, token=TOKEN, org=ORG)

BUCKET = 'test_bucket_12345'
START = 0
N_SAMPLES = 5
TEST_DATA_1 = {
    "measurement": 'testcopy',
    "device_id": 'testcopy.1',
    "values": [(datetime.now().replace(microsecond=0) - timedelta(minutes=i), i)
               for i in range(START, N_SAMPLES)]
}

TEST_DATA_2 = {
    "measurement": 'testcopy2',
    "device_id": 'testcopy.2',
    "values": [(datetime.now().replace(microsecond=0) - timedelta(minutes=i), i)
               for i in range(N_SAMPLES + 1, N_SAMPLES + N_SAMPLES)]
}


def test_copy_values():
    if CLIENT.buckets_api().bucket_exists(bucket_name=BUCKET):
        CLIENT.buckets_api().delete_bucket_by_name(bucket_name=BUCKET)
    CLIENT.buckets_api().create_bucket(bucket_name=BUCKET)

    CLIENT.delete_api().delete_data(bucket=BUCKET, measurements=['testcopy', 'testcopy2'])
    CLIENT.save_data(bucket=BUCKET, data=TEST_DATA_1)
    CLIENT.save_data(bucket=BUCKET, data=TEST_DATA_2)

    CLIENT.copy_from_to(bucket_from=BUCKET,
                        measurement_from=TEST_DATA_2['measurement'],
                        devid_from=TEST_DATA_2['device_id'],
                        bucket_to=BUCKET,
                        measurement_to=TEST_DATA_1['measurement'],
                        devid_to=TEST_DATA_1['device_id'])

    data_merged = CLIENT.get_data_from_device_id(bucket=BUCKET,
                                                 measurement=TEST_DATA_1['measurement'],
                                                 device_id=TEST_DATA_1['device_id'])
    for i, data in  enumerate(TEST_DATA_1['values'] + TEST_DATA_2['values']):
        assert str(data[0]) == data_merged['values'][i][0]
        assert data[1] == data_merged['values'][i][1]



from datetime import datetime
from typing import List

from influxdb_client import ChecksService, LesserThreshold, CheckStatusLevel, ThresholdCheck, DashboardQuery, \
    QueryEditMode, TaskStatusType, DeadmanCheck
from influxdb_client.rest import ApiException

from dms_influx2.exceptions import CheckError
from dms_influx2.utils import timestamp_to_influx_string, localize_dt


class ChecksApi(ChecksService):
    def __init__(self, client):
        self.org = client.org
        self.org_id = client.organizations_api().get_id_organization_by_name(self.org)
        super().__init__(api_client=client.api_client)

    def get_all_checks(self):
        return self.get_checks(org_id=self.org_id)

    def create_threshold_check(self,
                               check_name,
                               bucket_name,
                               measurement,
                               device_id,
                               thresholds: List,
                               check_id=None,
                               **kwargs):
        now = timestamp_to_influx_string(localize_dt(datetime.utcnow()))
        query = f'''
            option now = () => {now}
            from(bucket: "{bucket_name}")
              |> range(start: v.timeRangeStart, stop: v.timeRangeStop)
              |> filter(fn: (r) => r["_measurement"] == "{measurement}")
              |> filter(fn: (r) => r["device_id"] == "{device_id}")
              |> filter(fn: (r) => r["_field"] == "value")
              |> aggregateWindow(every: v.windowPeriod, fn: mean, createEmpty: false)
              |> yield(name: "mean")
        '''

        # Due to the bug in influx api we cannot post '0' as value for any of the thresholds
        # Easy fix is to replace 0 with 0.0001 which is almost like zero. Nobody will know the difference
        for threshold in thresholds:
            for key, val in threshold.items():
                try:
                    if float(val) == float(0):
                        threshold[key] = 0.0001
                except ValueError:
                    pass

        check = ThresholdCheck(name=f"{check_name}",
                               status_message_template="The value is on: ${ r._level } level!",
                               every="5s",
                               offset="0s",
                               query=DashboardQuery(edit_mode=QueryEditMode.ADVANCED, text=query),
                               thresholds=thresholds,
                               org_id=self.org_id,
                               status=TaskStatusType.ACTIVE)
        try:
            if check_id is not None:
                return self.put_checks_id(check_id=check_id, check=check)
            else:
                return self.create_check(check)
        except ApiException as e:
            if e.status == 422:
                raise CheckError(f"Check with name {check_name} already exists.")

    def create_deadmancheck(self,
                            check_name: str,
                            bucket_name: str,
                            measurement: str,
                            device_id: str,
                            time_since: str,
                            stale_time: str,
                            level: str,
                            check_id=None,
                            **kwargs):
        """
        Trigger when values are not reporting for <time_since> and stop checking after <stale_time>.
        Set status to <level>.
        if check_id is passed than make an update

        :param time_since: 90s or 120m, or 2h
        :param stale_time: 90s or 120m, or 2h
        :param level: CRIT, INFO, WARN, OK
        :return: created check
        """

        now = timestamp_to_influx_string(localize_dt(datetime.utcnow()))
        query = f'''
            option now = () => {now}
            from(bucket: "{bucket_name}")
              |> range(start: v.timeRangeStart, stop: v.timeRangeStop)
              |> filter(fn: (r) => r["_measurement"] == "{measurement}")
              |> filter(fn: (r) => r["device_id"] == "{device_id}")
              |> filter(fn: (r) => r["_field"] == "value")
              |> aggregateWindow(every: v.windowPeriod, fn: mean, createEmpty: false)
              |> yield(name: "mean")
        '''
        check = DeadmanCheck(name=f"{check_name}",
                             status_message_template="The value is on: ${ r._level } level!",
                             every="5s",
                             offset="0s",
                             query=DashboardQuery(edit_mode=QueryEditMode.ADVANCED, text=query),
                             time_since=time_since,
                             stale_time=stale_time,
                             org_id=self.org_id,
                             level=level,
                             status=TaskStatusType.ACTIVE)
        try:
            if check_id is not None:
                return self.put_checks_id(check_id=check_id, check=check)
            else:
                return self.create_check(check)
        except ApiException as e:
            if e.status == 422:
                raise CheckError(f"Check with name {check_name} already exists.")

    def check_id_exists(self, check_id):
        try:
            self.get_checks_id(check_id=check_id)
            return True
        except ApiException as e:
            if e.status == 404 or e.status == 400:
                return False



import json
from datetime import datetime
from typing import Union, Literal

from dms_influx2.utils import timestamp_to_influx_string, localize_dt


def get_flux_query(bucket=None,
                   measurement=None,
                   device_id=None,
                   description=None,
                   time_range='1h',
                   time_from=None,
                   time_to=None,
                   aggregate_window=None,
                   aggregate_func=None,
                   aggregate_create_empty=False,
                   sort=None,
                   count=None,
                   last=False,
                   first=False,
                   limit=None,
                   drop_columns=None,
                   keep_only_columns=None,
                   group_columns=None,
                   distinct=False,
                   field='value',
                   bucket_to=None,
                   org_to=None,
                   time_shift=None,
                   time_offset: int = None,
                   window_count: str = None,
                   value_type: str = None,
                   value: float = None,
                   value_within: bool = None,
                   value_min: float = None,
                   value_max: float = None,
                   errors_filter: Literal['keep', 'drop', 'only'] = 'drop') -> str:
    """

    :param first:
    :param window_count:
    :param time_offset:
    :param time_shift:
    :param field:
    :param bucket: Name of the bucket
    :param measurement: Name of the measurement
    :param device_id: ID of device
    :param description: Description of data set
    :param time_range:
    :param time_from:
    :param time_to:
    :param aggregate_window:
    :param aggregate_func:
    :param aggregate_create_empty:
    :param sort:
    :param count:
    :param last:
    :param limit:
    :param drop_columns:
    :param keep_only_columns:
    :param group_columns:
    :param distinct:
    :param bucket_to:
    :param org_to:
    :param value_type:
    :param value_within:
    :param value_min:
    :param value_max:
    :param value:
    :param errors_filter: Errors are all values == 55M; Keep errors, drop values, query only errors
    :return:
    """

    if bucket is None:
        raise ValueError("Invalid value for `bucket`, must not be `None`")

    stop = None

    if time_range == 'all':
        time_range = '100y'

    if time_range is not None:
        start = f"-{time_range}"
    else:
        start = "-1h"

    if time_from is not None:
        start = timestamp_to_influx_string(time_from, time_offset)

    if time_to is not None:
        stop = timestamp_to_influx_string(time_to, time_offset)

    if field is None:
        field = 'value'

    if time_shift is not None:
        now = timestamp_to_influx_string(datetime.utcnow())
    else:
        now = timestamp_to_influx_string(localize_dt(datetime.utcnow()))

    # Set now if local timestamp is used
    query_str = f'option now = () => {now}'

    query_str += f'\n\tfrom(bucket:"{bucket}")'
    if stop:
        query_str += f'''\n\t|> range(start: {start}, stop:{stop})'''
    else:
        query_str += f'''\n\t|> range(start: {start})'''

    if measurement:
        query_str += f'''\n\t|> filter(fn: (r) => r["_measurement"] == "{measurement}")'''

    if device_id:
        query_str += f'''\n\t|> filter(fn: (r) => r["device_id"] == "{device_id}")'''

    if description:
        query_str += f'''\n\t|> filter(fn: (r) => r["description"] == "{description}")'''

    if errors_filter == 'keep':
        pass
    elif errors_filter == 'drop':
        query_str += f'''\n\t|> filter(fn: (r) => r["_value"] != 55000000)'''
    elif errors_filter == 'only':
        query_str += f'''\n\t|> filter(fn: (r) => r["_value"] == 55000000)'''

    if value_type:
        if value_type == 'lesser' or value_type == 'greater' and value is not None:
            sign = '<' if value_type == 'lesser' else '>'
            query_str += f'''\n\t|> filter(fn: (r) => r["_value"] {sign} {value})'''
        if value_type == 'range':
            if value_within:
                query_str += f'''\n\t|> filter(fn: (r) => r["_value"] > {value_min} and r["_value"] < {value_max})'''
            else:
                query_str += f'''\n\t|> filter(fn: (r) => r["_value"] < {value_min} or r["_value"] > {value_max})'''

    query_str += f'''\n\t|> filter(fn: (r) => r["_field"] == "{field}")'''

    if aggregate_window:
        if type(aggregate_window) != str:
            raise ValueError("Invalid type for `aggregate_window`, must be type `str`")
        if aggregate_func is None:
            aggregate_func = "mean"
        create_empty = 'false'
        if aggregate_create_empty:
            create_empty = 'true'
        query_str += f'''\n\t|> aggregateWindow(every: {aggregate_window}, fn: {aggregate_func}, createEmpty: {create_empty})'''

    if window_count is not None:
        if type(window_count) != str:
            raise ValueError("Invalid type for `window_count`, must be type `str`")
        query_str += f'''\n\t|> window(every: {window_count}, createEmpty: true)'''
        query_str += f'''\n\t|> count()'''
        query_str += f'''\n\t|> rename(columns: ''' + '{_stop: "_time"})'
        query_str += f'''\n\t|> group(columns: ["_measurement", "device_id"], mode:"by")'''

    if group_columns:
        if type(group_columns) != list:
            raise ValueError("Invalid type for `group_columns`, must be type `list`")
        query_str += f'''\n\t|> group(columns: {json.dumps(group_columns)})'''

    if drop_columns:
        if type(drop_columns) != list:
            raise ValueError("Invalid type for `drop_columns`, must be type `list`")
        query_str += f'''\n\t|> drop(columns: {json.dumps(drop_columns)})'''

    if keep_only_columns:
        if type(keep_only_columns) != list:
            raise ValueError("Invalid type for `keep_only`, must be type `list`")
        query_str += f'''\n\t|> keep(columns: {json.dumps(keep_only_columns)})'''

    if count:
        query_str += f'''\n\t|> count(column: "_value")'''

    if sort is not None:
        # TODO: better sorting system
        # It appears that sort is very slow and cpu intensive
        if sort not in ['asc', 'desc']:
            raise ValueError("Invalid value for `sort`, must not be `asc` or `desc`")
        desc = "true" if sort == 'desc' else "false"
        query_str += f'''\n\t|> sort(columns: ["_time"], desc: {desc})'''

    if last:
        query_str += '''\n\t|> last()'''

    if first:
        query_str += '''\n\t|> first()'''

    if limit is not None:
        query_str += f'''\n\t|> limit(n:{limit}, offset: 0)'''

    if distinct:
        query_str += f'''\n\t|> distinct()'''

    if time_shift is not None:
        query_str += f'''\n\t|> timeShift(duration: {time_shift})'''

    if bucket_to is not None and org_to is not None:
        query_str += f'''\n\t|> to(bucket: "{bucket_to}", org: "{org_to}")'''
    return query_str



from influxdb_client import OrganizationsApi, Organization

from dms_influx2.exceptions import OrganizationsApiError


class Organizations(OrganizationsApi):
    def __init__(self, client):
        super().__init__(client)

    def get_id_organization_by_name(self, org_name: str) -> str:
        org_id = list(filter(lambda x: x.name == org_name, self.find_organizations()))
        if not org_id:
            raise OrganizationsApiError(f"Organisation `{org_name}` does not exists")
        else:
            return org_id[0].id

    def get_organization(self, org_name: str) -> Organization:
        orgs = self.find_organizations(org=org_name)
        return orgs[0] if orgs else None

    def get_organization_by_name(self, org_name: str):
        org_id = list(filter(lambda x: x.name == org_name, self.find_organizations()))
        if not org_id:
            raise OrganizationsApiError(f"Organisation `{org_name}` does not exists")
        else:
            return org_id[0]




import itertools
import logging
from datetime import datetime, timedelta
from typing import Union

from dateutil.parser import parse
from influxdb_client import DeleteApi

from dms_influx2.utils import timestamp_to_influx_string

logger = logging.getLogger(__name__)


class Delete(DeleteApi):
    def __init__(self, client):
        self.org = client.org
        self.time_offset = client.time_offset
        self.query_str = client.query_str
        super().__init__(client)

    def delete_data(self,
                    bucket,
                    measurements=None,
                    device_ids=None,
                    descriptions=None,
                    org: str = None,
                    time_range: str = None,
                    time_from: Union[str, datetime] = None,
                    time_to: Union[str, datetime] = None) -> dict:

        # TODO: add time_range
        if time_range is not None:
            pass

        if org is None:
            org = self.org

        if time_from is None:
            start = "1900-01-01T00:00:00Z"
        else:
            start = timestamp_to_influx_string(time_from, self.time_offset)

        if time_to is None:
            # Add some safety timedelta
            stop_time = datetime.now() + timedelta(days=1)
            stop = timestamp_to_influx_string(stop_time, self.time_offset)
        else:
            stop = timestamp_to_influx_string(time_to, self.time_offset)

        predicates = []

        if measurements is not None:
            _predicates = []
            for measurement in measurements:
                _predicates.append(f'_measurement="{measurement}"')
            predicates.append(_predicates)

        if device_ids is not None:
            _predicates = []
            for device_id in device_ids:
                _predicates.append(f'device_id="{device_id}"')
            predicates.append(_predicates)

        if descriptions is not None:
            _predicates = []
            for description in descriptions:
                _predicates.append(f'description="{description}"')
            predicates.append(_predicates)

        # Filter empty lists and find all combinations for predicates, execute all queries
        predicates = list(filter(None, predicates))
        predicates = [predicate for predicate in itertools.product(*predicates)]
        predicates = [" and ".join(i for i in predicate) for predicate in predicates]
        predicates = list(dict.fromkeys(predicates))
        for predicate in predicates:
            self.delete(start=parse(start), stop=parse(stop), predicate=predicate, bucket=bucket, org=org)
            logger.debug(f"Delete data {parse(start)}->{parse(stop)} from bucket: {bucket}, predicate: {predicate}")
        return {"start": start, "stop": stop, "predicates": predicates}



class BucketApiError(Exception):
    """Some problem with Bucket Api"""


class OrganizationsApiError(Exception):
    """Some problem with Organizations Api"""


class CheckError(Exception):
    """Some problem with checks"""


class NotificationError(Exception):
    """Some problem with notification service"""



from functools import wraps
from time import time
import logging

logger = logging.getLogger('runtime')


def runtime(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        start = time()
        f = func(self, *args, **kwargs)
        runtime = time() - start
        logger.debug(f'func:{func.__name__}, runtime:{round(runtime, 2)}')
        # if runtime > 0.1:
        #     logger.info(f'func:{func.__name__}, runtime:{round(time() - start, 2)}')
        # else:
        #     logger.debug(f'func:{func.__name__}, runtime:{round(time() - start, 2)}')
        return f

    return wrapper



from influxdb_client import BucketsApi, Bucket, Authorization

from dms_influx2.exceptions import BucketApiError


class Buckets(BucketsApi):
    def __init__(self, client):
        self.org = client.org
        self.authorizations_api = client.authorizations_api
        self.organizations_api = client.organizations_api
        super().__init__(client)

    def _create_bucket(self, bucket_name: str, description: str = None, org_name: str = None) -> Bucket:
        if org_name is None:
            org_name = self.org
        org = self.organizations_api().get_organization(org_name=org_name)
        return self.create_bucket(bucket_name=bucket_name, org=org, description=description)

    def bucket_exists(self, bucket_name):
        return bool(self.find_bucket_by_name(bucket_name))

    def create_bucket_with_auth(self, bucket_name: str, description: str = None,
                                org_name: str = None) -> Authorization:
        """
        Create new bucket with authorization
        :param bucket_name: bucket name
        :param description: bucket description
        :param org_name: organization name
        :return: token
        """

        if org_name is None:
            org_name = self.org

        org = self.organizations_api().get_organization(org_name=org_name)
        bucket = self._create_bucket(bucket_name=bucket_name, description=description, org_name=org_name)
        permissions_read = {
            'action': 'read',
            'resource': {
                'id': bucket.id,
                'name': bucket.name,
                'org': org.name,
                'org_id': org.id,
                'type': 'buckets'
            }
        }
        permissions_write = {
            'action': 'write',
            'resource': {
                'id': bucket.id,
                'name': bucket.name,
                'org': org.name,
                'org_id': org.id,
                'type': 'buckets'
            }
        }
        permissions = [permissions_read, permissions_write]
        authorization = Authorization(org_id=org.id, permissions=permissions, description=description)
        return self.authorizations_api().create_authorization(authorization=authorization)

    def delete_permissions(self, bucket_name, org=None, read=True, write=True, delete_if_none=True):
        if org is None:
            org = self.org

        organization = self.organizations_api().get_organization_by_name(org_name=org)
        auths = self.authorizations_api().find_authorizations_by_org(organization)
        for auth in auths:
            orig_perm = list(auth.permissions)
            for perm in auth.permissions:
                if locals()[perm.action]:
                    if perm.resource.name == bucket_name:
                        perm.resource.name = None
                        # auth.permissions[perm]
                        # auth.permissions.remove(perm)
            if orig_perm != auth.permissions:
                self.authorizations_api().update_authorization(auth=auth)
            if delete_if_none:
                if all([bool(i.resource.name is None and i.resource.id is not None) for i in auth.permissions]):
                    self.authorizations_api().delete_authorization(auth=auth)

    def delete_bucket_by_name(self, bucket_name: str):
        bucket = self.find_bucket_by_name(bucket_name=bucket_name)
        if bucket is None:
            raise BucketApiError(f"Bucket {bucket_name} does not exists.")
        self.delete_bucket(bucket=bucket)

    def delete_bucket_by_id(self, bucket_id):
        bucket = self.find_bucket_by_id(bucket_id)
        if bucket is None:
            raise BucketApiError(f"Bucket {bucket_id} does not exists.")
        self.delete_bucket(bucket=bucket)

    def list_buckets(self, only_names=False):
        buckets = []
        offset = 0
        while True:
            _buckets = self.find_buckets(offset=offset, limit=100).buckets
            if not _buckets:
                break
            buckets += _buckets
            offset += 100
        if only_names:
            return [bucket.name for bucket in buckets]
        else:
            return buckets

    def get_buckets_dict(self, order_by_created=True):
        buckets = [bucket.to_dict() for bucket in self.list_buckets()]
        if order_by_created:
            buckets = sorted(buckets, key=lambda b: b['created_at'], reverse=True)
        # buckets_out
        # for bucket in buckets:
        #     created_at = bucket['created_at']
        #     bucket['created_at'] = bucket['created_at'].tzinfo
        #     datetime(2021, 12, 6, 13, 33, 10, 146344).strftime("%Y-%m-%d %H:%M:%S")
        return buckets

    def bucket_exists_by_id(self, bucket_id):
        return bool(self.find_bucket_by_id(id=bucket_id))

    def change_bucket_description(self, bucket_id, description):
        bucket = self.find_bucket_by_id(bucket_id)
        bucket.description = description
        return self.update_bucket(bucket)


    def create_bucket_with_default_authorizations(self, bucket_name, bucket_description=None):
        """
        Create bucket with two authorizations:
        - read/write auth with description == bucket_name
        - read auth with description == bucket_name_readonyl

        """
        if bucket_description is None:
            bucket_description = bucket_name
        bucket = self._create_bucket(bucket_name=bucket_name, description=bucket_description)

        # read/write
        self.authorizations_api().create_bucket_authorization(bucket_name=bucket_name,
                                                              read_only=False,
                                                              description=bucket_name)

        # read
        self.authorizations_api().create_bucket_authorization(bucket_name=bucket_name,
                                                              read_only=True,
                                                              description=bucket_name + '_readonly')

        return bucket

    def create_bucket_if_not_exists(self, bucket_name):
        if not self.bucket_exists(bucket_name=bucket_name):
            self.create_bucket(bucket_name=bucket_name)




import os

from dms_influx2.lib import DmsInflux2

DATABASE_URL = os.environ['DATABASE_URL']
TOKEN = os.environ['TOKEN']
ORG = os.environ['ORG']
BUCKET = 'test_bucket'
CLIENT = DmsInflux2(url=DATABASE_URL, token=TOKEN, org=ORG)

BUCKET_TEST = 'test_buckets'
BUCKET_TEST_DESCRIPTION = 'test_buckets'
BUCKET_TEST_DESCRIPTION_READONLY = 'test_buckets_readonly'
AUTH_BUCKET = 'test2'


def test_bucket_exists():
    CLIENT.authorizations_api().delete_all_bucket_authorizations()

    if CLIENT.buckets_api().bucket_exists(bucket_name=BUCKET_TEST):
        CLIENT.buckets_api().delete_bucket_by_name(bucket_name=BUCKET_TEST)

    if CLIENT.buckets_api().bucket_exists(bucket_name=AUTH_BUCKET):
        CLIENT.buckets_api().delete_bucket_by_name(bucket_name=AUTH_BUCKET)

    CLIENT.buckets_api()._create_bucket(bucket_name=BUCKET_TEST)
    assert CLIENT.buckets_api().bucket_exists(BUCKET_TEST)


def test_change_bucket_description():
    pass


def test_create_bucket_with_default_authorizations():
    CLIENT.buckets_api().delete_bucket_by_name(bucket_name=BUCKET_TEST)
    CLIENT.buckets_api().create_bucket_with_default_authorizations(BUCKET_TEST)

    auth_read_write = CLIENT.authorizations_api().find_authorizations_by_description(BUCKET_TEST)[0]
    assert auth_read_write.description == BUCKET_TEST
    assert len(auth_read_write.permissions) == 2
    assert auth_read_write.permissions[0].action == 'read'
    assert auth_read_write.permissions[0].resource.type == 'buckets'
    assert auth_read_write.permissions[1].action == 'write'
    assert auth_read_write.permissions[1].resource.type == 'buckets'

    auth_read = CLIENT.authorizations_api().find_authorizations_by_description(BUCKET_TEST_DESCRIPTION_READONLY)[0]
    assert auth_read.description == BUCKET_TEST_DESCRIPTION_READONLY
    assert len(auth_read.permissions) == 1
    assert auth_read.permissions[0].action == 'read'
    assert auth_read.permissions[0].resource.type == 'buckets'

    ### Remove this permissions and test get_or_create_bucket_authorization()
    print('auth read', auth_read)
    CLIENT.authorizations_api().delete_authorization(auth_read_write)
    CLIENT.authorizations_api().delete_authorization(auth_read)

    auth = CLIENT.authorizations_api().get_or_create_bucket_authorization(bucket_name=BUCKET_TEST, read_only=True)
    assert auth.description == BUCKET_TEST_DESCRIPTION_READONLY
    assert len(auth.permissions) == 1
    assert auth.permissions[0].action == 'read'
    assert auth.permissions[0].resource.type == 'buckets'

    auth = CLIENT.authorizations_api().get_or_create_bucket_authorization(bucket_name=BUCKET_TEST, read_only=False)
    print(auth)
    assert auth.description == BUCKET_TEST_DESCRIPTION
    assert len(auth.permissions) == 2
    assert auth.permissions[0].action == 'read'
    assert auth.permissions[0].resource.type == 'buckets'
    assert auth.permissions[1].action == 'write'
    assert auth.permissions[1].resource.type == 'buckets'

def test_auths_dict():
    auths = CLIENT.authorizations_api().get_bucket_authorizations_dict()
    print(auths)
    assert True

def test_buckets_dict():
    pass

# # # # TODO fix this test
# def test_create_bucket_with_token_and_delete_permissions():
#     """Test create new bucket with auth token and check for writes/reads permissions"""
#
#     new_token = client.buckets_api().create_bucket_with_auth(bucket_name=bucket).token
#
#     measurement = "test"
#
#     client2 = DmsInflux2(url=url, token=new_token, org=org)
#     data = {
#         "measurement": measurement,
#         "device_id": "test.ch",
#         "values": [(datetime.utcnow(), 0)]
#     }
#     client2.save_data(bucket=bucket, data=data)
#     client.delete_api().delete_data(bucket=bucket, measurements=[measurement])
#
#     client.buckets_api().delete_permissions(bucket_name=bucket)
#     data = {
#         "measurement": measurement,
#         "device_id": "test.ch",
#         "values": [(datetime.utcnow(), 0)]
#     }
#     with pytest.raises(Exception):
#         client2.save_data(bucket=bucket, data=data)


def test_delete_bucket_by_name():
    if CLIENT.buckets_api().bucket_exists(bucket_name=BUCKET_TEST):
        CLIENT.buckets_api().delete_bucket_by_name(bucket_name=BUCKET_TEST)

    if CLIENT.buckets_api().bucket_exists(bucket_name=AUTH_BUCKET):
        CLIENT.buckets_api().delete_bucket_by_name(bucket_name=AUTH_BUCKET)


    assert not CLIENT.buckets_api().bucket_exists(BUCKET_TEST)
    assert not CLIENT.buckets_api().bucket_exists(AUTH_BUCKET)



import logging
import math
from time import sleep, time
from typing import List, Union, Literal

from ciso8601 import parse_datetime
from dateutil.parser import parse
from influxdb_client import InfluxDBClient, Point
from datetime import datetime, timedelta
from influxdb_client.client.flux_table import FluxTable
from influxdb_client.client.write_api import WriteOptions

from dms_influx2.authorizations import Authorizations
from dms_influx2.buckets import Buckets
from dms_influx2.checks import ChecksApi
from dms_influx2.decorators import runtime
from dms_influx2.delete import Delete
from dms_influx2.dtv_files import parse_dtv_bytes
from dms_influx2.notifications import NotificationEndpointApi, NotificationRuleApi
from dms_influx2.organisations import Organizations
from dms_influx2.query import get_flux_query
from dms_influx2.tasks import Tasks

from dms_influx2.utils import localize_dt, timestamp_to_influx_string

logger = logging.getLogger(__name__)


class DmsInflux2(InfluxDBClient):
    def __init__(self,
                 url=None,
                 token=None,
                 org=None,
                 enable_gzip=True,
                 timezone_offset: int = None,
                 timeout=20_000,
                 timezone='Europe/Ljubljana'):
        self.query_str = None
        self.predicates = None

        self.time_offset = timezone_offset
        self.time_shift = f'{timezone_offset}h' if timezone_offset is not None else None

        self.timezone = timezone
        self.downsample_temp_bucket = 'downsampled_temp_bucket'

        # Write params, modify for optimum performance
        # More info influxdb lib WriteOptions
        self.write_batch_size = 5000
        self.write_flush_interval = 1_000
        self.write_jitter_interval = 1_000
        self.write_retry_interval = 1_000
        self.write_max_retries = 5
        self.write_max_retry_delay = 30_000
        self.write_exponential_base = 2

        super().__init__(url=url, token=token, org=org, enable_gzip=enable_gzip, timeout=timeout)

    def buckets_api(self) -> Buckets:
        return Buckets(self)

    def organizations_api(self) -> Organizations:
        return Organizations(self)

    def delete_api(self) -> Delete:
        return Delete(self)

    def authorizations_api(self) -> Authorizations:
        return Authorizations(self)

    def checks_api(self):
        return ChecksApi(self)

    def notification_endpoint_api(self):
        return NotificationEndpointApi(self)

    def notification_rule_api(self):
        return NotificationRuleApi(self)

    def tasks_api(self) -> Tasks:
        return Tasks(self)

    def _list_tag_values(self, bucket, measurement, tag, **kwargs):
        query = f'import "influxdata/influxdb/schema"\n\nschema.measurementTagValues(' \
                f'bucket: "{bucket}",' \
                f'measurement: "{measurement}",' \
                f'tag: "{tag}")'
        self.query_str = query
        tables = self.query_api().query(query)
        data = []
        for table in tables:
            for record in table.records:
                data.append(record.get_value())
        return data

    def _list_distinct_tags(self, bucket, measurement, tag='description', time_range=None):

        if time_range is None:
            time_range = 'all'

        tables = self._get_tables(bucket=bucket,
                                  measurement=measurement,
                                  time_range=time_range,
                                  last=True,
                                  group_columns=[tag],
                                  distinct=True)
        return self._convert_tables_key(tables, key=tag)

    def _perform_query(self, query, params=None):
        query_api = self.query_api()
        self.query_str = query
        # print(query)
        return query_api.query(query, params=params)

    def _convert_tables_key(self, tables, key='description'):
        data = []
        for table in tables:
            for record in table.records:
                value = record.values.get(key)
                if value:
                    data.append(value)
        return data

    @runtime
    def _extract_tables(self, tables: List['FluxTable'], values_name='values', timestamp_to_string=True) -> list:

        data = []
        for table in tables:
            labels = [i.label for i in table.columns]
            try:
                get_columns = list(filter(lambda x: (x[0] != '_' and x not in ['table', 'result']), labels))
            except Exception as e:
                # TODO try to resolve this error
                for record in table.records:
                    logger.error(record)
                continue
            _data = {}
            _values = []
            for record in table.records:
                if not _data:
                    for col in get_columns:
                        _data[col] = record.values.get(col, None)
                dt = record.get_time()
                if timestamp_to_string:
                    dt = str(dt).split('+')[0]
                _values.append((dt, record.get_value()))
            data.append({**_data, values_name: _values})
        return data

    @runtime
    def _get_tables(self, **kwargs) -> List['FluxTable']:
        """Return tables based on flux query"""
        query = get_flux_query(**kwargs, time_shift=self.time_shift, time_offset=self.time_offset)
        return self._perform_query(query=query)

    def list_measurements(self, bucket) -> list:
        """Return list of measurements inside a bucket"""

        query = f'import "influxdata/influxdb/schema"\n\nschema.measurements(bucket: "{bucket}")'
        self.query_str = query
        tables = self.query_api().query(query)
        data = []
        for table in tables:
            for record in table.records:
                data.append(record.get_value())
        return data

    def list_device_ids(self, bucket, measurement=None) -> list:
        """Return list of distinct device_ids (tag=device_id)"""

        return self._list_distinct_tags(bucket=bucket, measurement=measurement, tag='device_id')

    def list_descriptions(self, bucket, measurement=None, time_range=None) -> list:
        """Return list of distinct descriptions"""

        if time_range is None:
            time_range = '100y'

        predicate = f'r._measurement == "{measurement}"' if measurement else 'true'

        query = f'''import "influxdata/influxdb/schema"\n\n
        schema.tagValues(
          bucket: "{bucket}",
          tag: "description",
          predicate: (r) => {predicate},
          start: -{time_range}
        )'''

        tables = self._perform_query(query=query)
        data = []
        for table in tables:
            for record in table.records:
                data.append(record.get_value())
        return data

    def get_values_count_combined(self, bucket, measurement=None, device_id=None, description=None,
                                  time_range=None, time_from=None, time_to=None,
                                  aggregate_window=None, aggregate_func=None) -> int:

        if time_range is None:
            time_range = 'all'

        tables = self._get_tables(bucket=bucket, measurement=measurement, device_id=device_id, description=description,
                                  time_range=time_range, time_from=time_from, time_to=time_to,
                                  aggregate_window=aggregate_window, aggregate_func=aggregate_func, count=True,
                                  keep_only_columns=["_value"])

        values_len = None
        for table in tables:
            for record in table.records:
                values_len = record.get_value()
        return values_len

    def get_values_count(self, bucket, measurement=None, device_id=None, description=None,
                         time_range=None, time_from=None, time_to=None,
                         aggregate_window=None, aggregate_func=None, window_count=None) -> dict:

        if time_range is None:
            time_range = '100y'

        local_vars = locals()
        local_vars.pop('self')
        local_vars['count'] = True
        local_vars['group_columns'] = ["device_id"]
        local_vars['keep_only_columns'] = ["_value", "device_id"]

        tables = self._get_tables(**local_vars)

        data = {}
        for table in tables:
            for record in table.records:
                data[record.values.get("device_id")] = record.get_value()
        return data

    def get_metadata(self, bucket, measurement=None, device_id=None, description=None,
                     time_range=None, time_from=None, time_to=None) -> list:
        """Get len and last value"""

        local_vars = locals()
        local_vars.pop('self')

        data = self.get_one_value(**local_vars)
        data_len = self.get_values_count(**local_vars)

        for item in data:
            device_id = item['device_id']
            if device_id in data_len:
                item['values_count'] = data_len[device_id]
            else:
                item['values_count'] = 0

        return data

    @staticmethod
    def pop_timestamp_from_values(data):
        return data['values'][-1][0]

    @staticmethod
    def to_one_value_format(data):
        # Transform data from tables to one value format:
        _data = []
        for item in data:
            item['timestamp'] = item['values'][-1][0]
            item['value'] = item['values'][-1][1]
            item.pop('values')
            _data.append(item)
        return _data

    def get_one_value(self, bucket, measurement=None, device_id=None, description=None,
                      time_range=None, time_from=None, time_to=None, sort='desc', last=True,
                      errors_filter='keep') -> list:
        """Get only one value per table (last or first)"""

        if time_range is None and time_from is None and time_to is None:
            time_range = 'all'

        tables = self._get_tables(bucket=bucket, measurement=measurement, device_id=device_id, description=description,
                                  time_range=time_range, time_from=time_from, time_to=time_to, last=last,
                                  value_type=None, value=None, value_within=None, value_min=None, value_max=None,
                                  first=not last, group_columns=["device_id"], errors_filter=errors_filter)

        data = self._extract_tables(tables)

        if sort == 'desc' or sort == 'asc':
            data.sort(key=self.pop_timestamp_from_values, reverse=sort == 'desc')

        return self.to_one_value_format(data)

    def get_values_from_device_id(self, bucket, measurement, device_id, description=None,
                                  time_range=None, time_from=None, time_to=None,
                                  aggregate_window=None, aggregate_func=None, aggregate_create_empty=False,
                                  value_type=None, value=None, value_within=None, value_min=None, value_max=None,
                                  sort='desc', limit=None, window_count=None, errors_filter=None):

        if measurement is None:
            raise ValueError("Invalid value for `measurement`, must not be `None`")

        if device_id is None:
            raise ValueError("Invalid value for `device_id`, must not be `None`")

        local_vars = locals()
        local_vars.pop('self')
        local_vars['drop_columns'] = ["unit"]
        local_vars['keep_only_columns'] = ["_value", "_time"]

        tables = self._get_tables(**local_vars)

        values = self._extract_tables(tables)

        return values[0]['values'] if values else None

    def get_data_from_device_id(self, bucket, measurement, device_id, description=None,
                                time_range=None, time_from=None, time_to=None,
                                aggregate_window=None, aggregate_func=None, aggregate_create_empty=False,
                                value_type=None, value=None, value_within=None, value_min=None, value_max=None,
                                sort='desc', limit=None, window_count=None, errors_filter=False) -> dict:

        if measurement is None:
            raise ValueError("Invalid value for `measurement`, must not be `None`")

        if device_id is None:
            raise ValueError("Invalid value for `device_id`, must not be `None`")

        local_vars = locals()
        local_vars.pop('self')

        data = self.get_one_value(bucket=bucket,
                                  measurement=measurement,
                                  device_id=device_id,
                                  time_range=time_range,
                                  time_from=time_from,
                                  time_to=time_to,
                                  description=description)
        if data:
            data = data[0]
            data['values'] = self.get_values_from_device_id(**local_vars)
        else:
            data = {}
        return data

    def save_data(self, bucket, data, utc_to_local=False, timezone='Europe/Ljubljana', org=None, create_bucket=True):
        """ Save data to database

        sample_data = [{
            measurement: <str>,
            device_id: <str> (required),
            device: <str>,
            channel: <str>,
            description: <str>,
            unit: <str> (required),
            values: [[ <time>, value], ...] (required)
            timestamp: <time> (optional)
            value: <float> (optional)
        }]

        Note: timestamp must be saved without tzinfo, or it is removed

        :param bucket: Bucket
        :param data: Data to save (sample_data)
        :param org: Organisation
        :param utc_to_local: Convert time as utc to local time (timezone)
        :param timezone: Use timezone for utc conversion
        :param create_bucket: Create bucket if it does not exists
        """

        if create_bucket:
            self.buckets_api().create_bucket_if_not_exists(bucket_name=bucket)

        if type(data) == dict:
            data = [data]

        if org is None:
            org = self.org

        points = []
        for item in data:
            try:
                if not item['device_id']:
                    raise ValueError('Device id must be supplied')
                device_id = item['device_id']
                measurement = item.get('measurement', device_id.split('.')[0])
                device = item.get('device', measurement)
                channel = item.get('channel', device_id.split('.')[1])
                description = item.get('description', None)
                unit = item.get('unit', '')
                value = item.get('value', None)
                timestamp = item.get('timestamp', None)
                if 'values' not in item and timestamp is not None and value is not None:
                    item['values'] = [(timestamp, value)]

                for values in item['values']:
                    if type(values) not in [list, tuple]:
                        raise ValueError('Point must be `tuple` or `list` example: (2021-01-01 00:00:00, 1)')

                    dt = values[0]
                    value = float(values[1])

                    if type(dt) == str:
                        try:
                            dt = parse_datetime(dt)
                        except Exception:
                            logger.error(f"Invalid time string, ts:{dt}, data:{item}")
                            continue

                    if dt.tzinfo is not None:
                        dt = dt.replace(tzinfo=None)

                    if utc_to_local:
                        dt = localize_dt(dt)

                    if dt.year == 1970:
                        logger.error(f"Invalid timestamp error for year 1970, {measurement}:{device_id}")
                        continue

                    point = Point(measurement) \
                        .tag("device_id", device_id) \
                        .tag("device", device) \
                        .tag("channel", channel) \
                        .tag("unit", unit) \
                        .tag("timezone", timezone) \
                        .field("value", value) \
                        .time(time=dt)
                    if description is not None:
                        point.tag('description', description)
                    points.append(point)
            except Exception as e:
                logger.exception(e)
                logger.error(f"unable to write point, e:{e}")

        with self.write_api(write_options=WriteOptions(batch_size=self.write_batch_size,
                                                       flush_interval=self.write_flush_interval,
                                                       jitter_interval=self.write_jitter_interval,
                                                       retry_interval=self.write_retry_interval,
                                                       max_retries=self.write_max_retries,
                                                       max_retry_delay=self.write_max_retry_delay,
                                                       exponential_base=self.write_exponential_base)) as _write_client:
            _write_client.write(bucket=bucket, org=org, record=points)

        if points:
            logger.info(f"written {len(points)} to bucket: {bucket}")

    def copy_from_to(self, bucket_from, measurement_from, devid_from, bucket_to, measurement_to, devid_to):
        data_from = self.get_data_from_device_id(bucket=bucket_from,
                                                 measurement=measurement_from,
                                                 device_id=devid_from,
                                                 time_range='100y')
        data_from['measurement'] = measurement_to
        data_from['device'] = measurement_to
        data_from['device_id'] = devid_to
        data_from['channel'] = ".".join(devid_to.split('.')[1:])

        self.save_data(bucket=bucket_to, data=data_from)

    def copy_bucket_to_bucket(self, bucket_from, bucket_to):
        query = 'from(bucket:"{bucket_from}") |> range(start: -1h)'

        records = self.query_api().query_stream(query)

        for record in records:
            print(record)
            # print(f'Temperature in {record["location"]} is {record["_value"]}')

    def _copy_data(self, bucket_from, bucket_to, org_to=None, measurement=None, device_id=None, description=None,
                   time_range=None, time_from=None, time_to=None, aggregate_window=None, aggregate_func=None,
                   sort='desc', limit=None):
        """Copy data from one bucket to another bucket"""

        if org_to is None:
            org_to = self.org

        if time_range is None:
            time_range = 'all'

        query = get_flux_query(bucket=bucket_from, bucket_to=bucket_to, org_to=org_to, measurement=measurement,
                               device_id=device_id, description=description,
                               time_range=time_range, time_from=time_from, time_to=time_to,
                               aggregate_window=aggregate_window,
                               aggregate_func=aggregate_func,
                               sort=None, limit=limit)
        self._perform_query(query)

    def move_data_to_trash(self, bucket, measurement, bucket_trash=None, device_id=None, description=None,
                           time_range=None, time_from=None, time_to=None, delete_afterwords=True):

        if measurement is None:
            raise ValueError("Invalid value for `measurement`, must not be `None`")

        if bucket_trash is None:
            bucket_trash = f'{bucket}_trash_{datetime.now().year}'

        bucket_api = self.buckets_api()
        if not bucket_api.bucket_exists(bucket_name=bucket_trash):
            org = self.organizations_api().get_organization(org_name=self.org)
            bucket_api.create_bucket(bucket_name=bucket_trash, org=org,
                                     description=f"Trash bucket for year {datetime.now().year}")

        self._copy_data(bucket_from=bucket, bucket_to=bucket_trash, measurement=measurement, device_id=device_id,
                        description=description, time_range=time_range, time_from=time_from, time_to=time_to)

        if delete_afterwords:
            device_ids = [device_id] if device_id is not None else []
            descriptions = [description] if description is not None else []
            self.delete_api().delete_data(bucket=bucket, measurements=[measurement], device_ids=device_ids,
                                          descriptions=descriptions, org=self.org,
                                          time_from=time_from, time_to=time_to)

    def count_data_in_buckets(self):
        buckets = self.buckets_api().list_buckets(only_names=True)
        counts = {}
        for bucket in buckets:
            count = self.get_values_count_combined(bucket=bucket)
            if count is None:
                count = 0
            counts[bucket] = count
        return dict(sorted(counts.items(), key=lambda item: item[1]))

    def save_dtv_data(self, filepath, device_id, bucket):
        if device_id is None:
            raise ValueError("Argument `devid` must not be None.")
        with open(filepath, 'rb') as f:
            file = f.read()
            data = parse_dtv_bytes(file)
        data_write = {
            "measurement": device_id.split('.')[0],
            "device_id": device_id,
            "values": data['values'],
            "unit": data['unit']
        }
        self.save_data(bucket=bucket, data=data_write)

    def _get_time_from_aggregate_window(self,
                                        time_window: str,
                                        dt: datetime = datetime.utcnow(),
                                        operator: Literal['plus', 'minus'] = 'minus') -> datetime:
        window_type = time_window[-1]
        window_number = int(time_window[0:-1])
        if dt is None:
            dt = datetime.utcnow()
        if window_type == 'y':
            window_type = 'd'
            window_number = window_number * 365
        d = {
            "s": "seconds",
            "m": "minutes",
            "h": "hours",
            "d": "days"
        }
        if operator == 'plus':
            return dt + timedelta(**{d[window_type]: window_number})
        if operator == 'minus':
            return dt - timedelta(**{d[window_type]: window_number})

    def down_sample(self,
                    bucket: str,
                    aggregate_window: str,
                    measurement: str = None,
                    device_id: str = None,
                    time_from: Union[str, datetime] = None,
                    time_to: Union[str, datetime] = None,
                    older_than: str = None,
                    decimal_places: int = None,
                    use_utc: bool = False,
                    to_int: bool = False):

        # TODO what about 55.mio and dates with 1970

        if time_from is None and time_to is None and older_than is None:
            raise ValueError("Params (`time_from`, `time_to`, `older_than`) must not all be None")

        if type(time_from) == str:
            parse(time_from)
        if type(time_to) == str:
            parse(time_to)

        if decimal_places is None:
            decimal_places = 3

        if use_utc is None:
            use_utc = False

        if older_than is not None:
            if use_utc:
                dt = datetime.utcnow()
            else:
                dt = localize_dt(datetime.utcnow())
            time_to = self._get_time_from_aggregate_window(time_window=older_than,
                                                           dt=dt,
                                                           operator='minus')
            time_from = None

        if time_from is not None:
            start = timestamp_to_influx_string(time_from)
        else:
            start = timestamp_to_influx_string(datetime(1980, 1, 1, 0, 0, 0))

        if time_to is not None:
            stop = timestamp_to_influx_string(time_to)
        else:
            if use_utc:
                stop = timestamp_to_influx_string(datetime.utcnow())
            else:
                stop = timestamp_to_influx_string(localize_dt(datetime.utcnow()))

        # Erase bucket every each time for this operation
        downsampled_bucket = self.downsample_temp_bucket
        if self.buckets_api().bucket_exists(bucket_name=downsampled_bucket):
            self.buckets_api().delete_bucket_by_name(bucket_name=downsampled_bucket)
        self.buckets_api().create_bucket(bucket_name=downsampled_bucket)

        decimal_places_number = math.pow(10, decimal_places)

        query_str = 'import "math"'
        query_str += f'''\n\tfrom(bucket:"{bucket}") |> range(start: {start}, stop: {stop})'''
        if measurement is not None:
            query_str += f'''\n\t|> filter(fn: (r) => r["_measurement"] == "{measurement}")'''
        if device_id is not None:
            query_str += f'''\n\t|> filter(fn: (r) => r["device_id"] == "{device_id}")'''
        query_str += f'''\n\t|> filter(fn: (r) => r["_value"] != 55000000)'''
        query_str += f'''\n\t|> aggregateWindow(every: {aggregate_window}, fn: mean, createEmpty: false)'''
        query_str += f'''\n\t|> map(
                            fn: (r) =>
                            ({{r with _value: (math.round(x: r._value * {decimal_places_number}) / {decimal_places_number}),}}),
                        )'''
        if to_int:
            query_str += f'''\n\t|> toInt()'''
        query_str += f'''\n\t|> to(bucket: "{downsampled_bucket}", org: "{self.org}")'''

        self.query_api().query(query=query_str)

        if measurement:
            if device_id is None:
                predicates = self.delete_api().delete_data(bucket=bucket,
                                                           time_from=time_from,
                                                           time_to=time_to,
                                                           measurements=[measurement])
            else:
                predicates = self.delete_api().delete_data(bucket=bucket,
                                                           time_from=time_from,
                                                           time_to=time_to,
                                                           device_ids=[device_id],
                                                           measurements=[measurement])

        else:
            predicates = self.delete_api().delete_data(bucket=bucket,
                                                       time_from=time_from,
                                                       time_to=time_to)

        query = f'''
            from(bucket: "{downsampled_bucket}") |> range(start: -100y)
                |> to(bucket: "{bucket}", org: "{self.org}")
        '''
        self.query_api().query(query=query)

    def __get_query_telegraf_fields(self,
                                    bucket: str,
                                    measurement: str,
                                    start: str,
                                    stop: str,
                                    aggregate_window: str,
                                    fields: List[str],
                                    to_int: bool = False):

        query = 'import "math"'
        query += f'''\n\tfrom(bucket:"{bucket}") |> range(start: {start}, stop: {stop})'''
        query += f'''\n\t|> filter(fn: (r) => r["_measurement"] == "{measurement}")'''
        if fields is not None:
            first_field = fields[0]
            query += f'''\n\t|> filter(fn: (r) => r["_field"] == "{first_field}"'''
            if len(fields) >= 1:
                for field in fields[1:]:
                    query += f''' or r["_field"] == "{field}"'''
            query += f')'
        query += f'''\n\t|> aggregateWindow(every: {aggregate_window}, fn: mean, createEmpty: false)'''
        if to_int:
            query += f'''\n\t|> toInt()'''
        query += f'''\n\t|> to(bucket: "{self.downsample_temp_bucket}", org: "{self.org}")'''

        return query

    def _downsample_telegraf_fields(self,
                                    bucket: str,
                                    measurement: str,
                                    aggregate_window: str = '1h',
                                    older_than: str = '30d',
                                    normal_fields: List[str] = None,
                                    to_int_fields: List[str] = None):

        if measurement is None:
            raise ValueError('Invalid value for argument `measurement`, cannot be None')

        time_to = self._get_time_from_aggregate_window(time_window=older_than,
                                                       dt=datetime.utcnow(),
                                                       operator='minus')
        time_from = None

        # Erase bucket every each time for this operation
        downsampled_bucket = self.downsample_temp_bucket
        if self.buckets_api().bucket_exists(bucket_name=downsampled_bucket):
            self.buckets_api().delete_bucket_by_name(bucket_name=downsampled_bucket)
        self.buckets_api().create_bucket(bucket_name=downsampled_bucket)

        start = timestamp_to_influx_string(datetime(1980, 1, 1, 0, 0))
        stop = timestamp_to_influx_string(time_to)

        if normal_fields is not None:
            query_normal_fields = self.__get_query_telegraf_fields(bucket=bucket,
                                                                   measurement=measurement,
                                                                   aggregate_window=aggregate_window,
                                                                   start=start,
                                                                   stop=stop,
                                                                   fields=normal_fields,
                                                                   to_int=False)
            self.query_api().query(query=query_normal_fields)

        if to_int_fields is not None:
            query_to_int_fields = self.__get_query_telegraf_fields(bucket=bucket,
                                                                   measurement=measurement,
                                                                   aggregate_window=aggregate_window,
                                                                   start=start,
                                                                   stop=stop,
                                                                   fields=to_int_fields,
                                                                   to_int=True)
            self.query_api().query(query=query_to_int_fields)

        self.delete_api().delete_data(bucket=bucket,
                                      measurements=[measurement],
                                      time_to=time_to)

        query_copy_to_bucket = f'''
            from(bucket: "{downsampled_bucket}") |> range(start: -100y)
                |> to(bucket: "{bucket}", org: "{self.org}")
        '''
        self.query_api().query(query=query_copy_to_bucket)

    def downsample_telegraf_stats(self,
                                  bucket: str,
                                  aggregate_window: str = '1h',
                                  older_than: str = '30d'):
        """
        cpu-fields: usage_guest, usage_guest_nice, usage_idle, usage_iowait, usage_irq, usage_nice, usage_softirq,
            usage_steal, usage_system, usage_user
        disk-fields: free, inodes_free, inodes_total, inodes_used, total, used, used_percent
        kernel-fields: boot_time, context_switches, entropy_avail, interrupts, processes_forked
        mem-fields: active, available, available_percent, buffered, cached, commit_limit, committed_as, dirty, free,
            high_free, high_total, huge_page_size, huge_pages_free, huge_pages_total, inactive, low_free, low_total,
            mapped, page_tables, shared, slab, sreclaimable, sunreclaim, swap_cached, swap_free, swap_total, total,
            used, used_percent, vmalloc_chunk, vmalloc_total, vmalloc_used, write_back, write_back_tmp
        """
        if older_than is None:
            older_than = '30d'
        items = [
            {
                "measurement": "cpu",
                "normal_fields": ["usage_guest", "usage_guest_nice", "usage_idle", "usage_iowait", "usage_irq",
                                  "usage_nice", "usage_softirq", "usage_steal", "usage_system", "usage_user"],
                "to_int_fields": None
            },
            {
                "measurement": "disk",
                "normal_fields": ["used_percent"],
                "to_int_fields": ['free', 'inodes_free', 'inodes_total', 'inodes_used', "total", 'used']
            },
            {
                "measurement": "kernel",
                "normal_fields": None,
                "to_int_fields": ['boot_time', 'context_switches', 'entropy_avail', 'interrupts', 'processes_forked']
            },
            {
                "measurement": "mem",
                "normal_fields": ['available_percent', 'used_percent'],
                "to_int_fields": ["active", 'available', 'buffered', 'cached', 'commit_limit', 'committed_as', 'dirty',
                                  'free', 'high_free', 'high_total', 'huge_page_size', 'huge_pages_free',
                                  'huge_pages_total', 'inactive', 'low_free', 'low_total', 'mapped', 'page_tables',
                                  'shared', 'slab', 'sreclaimable', 'sunreclaim', 'swap_cached', 'swap_free',
                                  'swap_total', 'total', 'used', 'vmalloc_chunk', 'vmalloc_total', 'vmalloc_used',
                                  'write_back', 'write_back_tmp']
            }
        ]

        start = time()
        for item in items:
            try:
                self._downsample_telegraf_fields(bucket=bucket,
                                                 aggregate_window=aggregate_window,
                                                 measurement=item['measurement'],
                                                 older_than=older_than,
                                                 normal_fields=item['normal_fields'],
                                                 to_int_fields=item['to_int_fields'])
            except Exception as e:
                logger.error(f'Cannot downsample telegraf stats for: {item["measurement"]}, e:{e}')
        logger.info(f'Downsample complete, bucket: {bucket}, type: telegraf_stats, runtime:{round(time() - start, 2)}s')
